loadClass("core")

//========================================================================================================
// Global Info.

var DX, DY, MAX_VIEW, MAX_SPEED, MAX_ROT, MAX_HARD_ROT, INI_POS, OVERRUN, DELAY=170
var b_objs={}, c_objs={}, m_objs=[]

function updateObstacleInformation(objs)
{
	m_objs.length = 0
	for each(var o in objs)
		     if( o.tag == "m" ) m_objs.push(o)
		else if( o.tag == "b" ) b_objs[o.p] = o
		else if( o.tag == "c" ) c_objs[o.p] = o
}

//========================================================================================================
// Predictor :: Predict the position and the direction in the future

function predictLinear(p, s, d, t)
{
	return vAdd(p, vMul(vecP(s,d), t))
}

function predictCurrentPosition(vc, p, s, d, delay)
{
	var dd = {L:MAX_HARD_ROT, l:MAX_ROT, "-":0, r:-MAX_ROT, R:-MAX_HARD_ROT}[vc[1]]
	var p2 = predictLinear(p, s, d+dd/2, delay)
	p2.r = p.r
	return [p2, s, d+dd*delay] // TODO: take into account accelerations etc...
}

//========================================================================================================
// Directors :: Each of them independently gives an advice for the direction the rover should go

function D_TowardHome(NAME, c0, c1, c2) // Hurry to Home!
{
	return function(p, s, d) {
		var x = vLen(p)
		var mag = c0 + c1/x + c2/x/x
		///print(NAME,":",mag)
		return vMul(vUnit(p), -mag)
	}
}

function lineAvoider(NAME, o,x,p,w,c1,c2,c3,prefer_home_rate)
{
	x = vUnit(x)
	p = vSub(p, o)
	var y = vRot(x, Math.PI/2)
	var px = vIP(x, p)
	var py = vIP(y, p)
	if( py < -w || w < py )
		return vec(0,0)
	var ax = Math.abs(px)
	if( ax < 0.000000001 ) return vec(0,0)
	var mag = c1/ax + c2/ax/ax + c3/ax/ax/ax
	if( px <= 0 )
		mag = -mag/3
	var center = 0
	if( px > w*3 )
		center = - sign(vIP(y, vNeg(o)))*w*prefer_home_rate
	///print(NAME, ":", mag)
	return vMul(y, (py<center ? -1 : 1) * mag)
}

function D_AvoidForward(NAME, objs, c1, c2, c3) // Avoid the Obstacles forward
{
	return function(p, s, d) {
		var q = vec(0,0)
		for each(var o in objs)
			if( vDist(o.p, p) < Math.max(MAX_VIEW, s*3000) )
				q = vAdd(q, lineAvoider(NAME,o.p, vecP(1, d+Math.PI), p, 0.5+o.r*2, c1, c2, c3, 0.5))
		return q
	}
}

function D_AvoidMoving(NAME, objs, c1, c2, c3) // Avoid Moving Enemies
{
	return function(p, s, d) {
		var q = vec(0,0)
		for each(var o in objs) {
			var x = Math.max( vDist(p,o.p)-p.r-o.r, 0.000000000001 )
			var mag = c1/x + c2/x/x + c3/x/x/x
			///print(NAME, ":", mag)
			var qq = vMul(vUnit(vSub(p,o.p)), mag)  // -pi .. pi
			var ag = vArg(vComplexDiv(qq, vNeg(p)))
			q = vAdd(q, vRot(qq, (ag<Math.PI ? ag : ag-Math.PI*2)/4-ag)) // -pi/4 .. pi/4
		}
		return q
	}
}

var Directors = [
	D_TowardHome  ("* Hickie",      0, 200, 6000    ),
	D_AvoidForward("% Boomkn", b_objs,  60,    0,  0),
	D_AvoidForward("# Claire", c_objs, 120,    0,  0),
	D_AvoidMoving ("& Martin", m_objs,   0,  500,  0),
]

function calcDirectionToGo(p, s, d) // Then the rover simply takes the sum of the adiveces...
{
	var v = vec(0,0)
	for each(D in Directors)
		v = vAdd(v, D(p, s, d))
	return v
}

//========================================================================================================
// Adjustment

function adjustControl(dd)
{
	var a = "a", l = "-"
	if( Math.PI*1/2 <= dd && dd <= Math.PI*3/2 ) a = "-"
	if( Math.PI*2/3 <= dd && dd <= Math.PI*4/3 ) a = "b"
	if( dd < Math.PI/24 )         l = "-"
	else if( dd < Math.PI/6 )     l = "l"
	else if( dd < Math.PI )       l = "L"
	else if( dd < Math.PI*11/6 )  l = "R"
	else if( dd < Math.PI*47/24 ) l = "r"
	return a+l
}

//========================================================================================================
// Main Loop

core_main_loop(arguments, {
	I: function(I, send)
	{
		[DX,DY,MAX_VIEW,MAX_SPEED,MAX_ROT,MAX_HARD_ROT] = [I.dx,I.dy,I.Msen,I.Mspd,I.Mtrn,I.Mhtrn]
		send( "--", "a-" )
	},
	E: function(E, send)
	{
		INI_POS = OVERRUN = undefined
	},
	T: function(M, send)
	{
		updateObstacleInformation(M.objs) // Update Informations
		var [p, s, d] = predictCurrentPosition(M.vc, M.p, M.spd, M.d, DELAY) // position,speed,direction

		if( !INI_POS )          INI_POS = M.p  // Info. for 
		if( vIP(INI_POS, p)<0 ) OVERRUN = true // Overrun check

		var v  = calcDirectionToGo(p, s, d) // Which direction to Go?
		var dd = (vArg(v)-d + 10*Math.PI) % (2*Math.PI) // diffrence from the current direction

		var target_vc = adjustControl(dd) // Decide the Best Control
		if( OVERRUN && vLen(p)<30 ) {     // If overrunning the goal, stop
			var dov = (vArg(p)-d + 10*Math.PI) % (2*Math.PI)
			if( dov < Math.PI*3/4 || Math.PI*5/4<dov ) {
				target_vc = "b" + target_vc[1]
				///print("@ Oberon", ":", "BRAKE!!!!")
			}
		}

		send( M.vc, target_vc ) // Send
	},
})
