// float ops
function sign(x)     { return x>0 ? 1 : x<0 ? -1 : 0 }

// constructor
function vec(x, y)    { return [x, y] }
function vecP(r, tht) { return vMul(vec(Math.cos(tht), Math.sin(tht)), r) }
function vecR(x, y, r){ var v = vec(x,y); v.r = r; return v } // + radius

// add, sub
function vAdd(v1, v2) { return [v1[0]+v2[0], v1[1]+v2[1]] }
function vSub(v1, v2) { return [v1[0]-v2[0], v1[1]-v2[1]] }
function vNeg(v)      { return vMul(v, -1) }

// vec * scalar
function vMul(v, s)   { return [v[0]*s, v[1]*s] }
function vDiv(v, s)   { return [v[0]/s, v[1]/s] }

// inner product
function vIP(v1,v2)   { return v1[0]*v2[0] + v1[1]*v2[1] }

// length
function vNorm(v)     { return vIP(v,v) }
function vLen(v)      { return Math.sqrt(vNorm(v)) }
function vDist(v1,v2) { return vLen(vSub(v1,v2)) }

// unit vector
function vUnit(v)     { var l = vLen(v); return l>0 ? vDiv(v,l) : v }

// complex op
function vComplexMul(v1,v2) { return [v1[0]*v2[0]-v1[1]*v2[1], v1[0]*v2[1]+v1[1]*v2[0]] }
function vComplexInv(v)     { var n=vNorm(v); return [v[0]/n, -v[1]/n] }
function vComplexDiv(v1,v2) { return vComplexMul(v1, vComplexInv(v2)) }

// rotate
function vRot(v,tht)  { return vComplexMul(v, vecP(1,tht)) }

// projection
function vProj(v,dir) { return vMul(dir, vIP(v,dir)/vNorm(dir)) }

// arg (0-2pi)
function vArg(v) { var a = Math.atan2(v[1],v[0]); return a<0 ? a+2*Math.PI : a }
